def pretty_dict(d, indent=0):
    """Pretty the output format of a dictionary.

    Parameters
    ----------
    d
        dict, the input dictionary instance.
    indent
        int, indent level, non-negative.
    Returns
    -------
    res
        str, the output string
    """

    res = ""
    for k, v in d.items():
        res += "\t" * indent + str(k)
        if isinstance(v, dict):
            res += "\n" + pretty_dict(v, indent + 1)
        else:
            res += ": " + str(v) + "\n"
    return res


if __name__ == "__main__":
    print(pretty_dict({"123": 12, "wqe": {"asf": 23}}))



class Ordering(object):
    '''
    A class representing variable ordering.

    '''

    def __new__(cls, ordering):
        if isinstance(ordering, list):
            return ListOrdering(ordering)

        TypeError('unsupported ordering %s' % (ordering))

    def __eq__(self, ordering):
        raise RuntimeError('not implemented')

    def __ne__(self, ordering):
        return not (self == ordering)

    def in_order(self, x, y):
        raise RuntimeError('not implemented')

    def cmp(self, x, y):
        raise RuntimeError('not implemented')

    def __contains__(self, var):
        raise RuntimeError('not implemented')


class FunctionOrdering(Ordering):
    '''
    A class representing variable ordering specified by a function.

    '''

    def __new__(cls, ordering):
        return super(Ordering, cls).__new__(cls)

    def __init__(self, ordering):
        if not isinstance(ordering, function):
            TypeError('expected a function, got %s' % (ordering))

        self.ordering = ordering

    def __eq__(self, ordering):
        if not isinstance(ordering, FunctionalOrdering):
            return False

        return self.ordering is ordering.ordering

    def in_order(self, x, y):
        return self.ordering(x, y)

    def __contains__(self, var):
        return True

    def cmp(self, x, y):
        if x == y:
            return 0

        if self.in_order(x, y):
            return -1

        return 1

    def __str__(self):
        return '%s' % (self.ordering)


def cmp_to_key(mycmp):
    'Convert a cmp= function into a key= function'
    class K:
        def __init__(self, obj, *args):
            self.obj = obj
        def __lt__(self, other):
            return mycmp(self.obj, other.obj) < 0
        def __gt__(self, other):
            return mycmp(self.obj, other.obj) > 0
        def __eq__(self, other):
            return mycmp(self.obj, other.obj) == 0
        def __le__(self, other):
            return mycmp(self.obj, other.obj) <= 0
        def __ge__(self, other):
            return mycmp(self.obj, other.obj) >= 0
        def __ne__(self, other):
            return mycmp(self.obj, other.obj) != 0
    return K


class ListOrdering(Ordering):
    '''
    A class representing variable ordering specified by a list.

    '''

    def __new__(cls, ordering):
        return super(Ordering, cls).__new__(cls)

    def __init__(self, ordering):
        if not isinstance(ordering, list):
            TypeError('expected a list, got %s' % (ordering))

        self.ordering = dict()

        i = 0
        for obj in ordering:
            if obj in self.ordering:
                raise RuntimeError('%s is repeated in %s' % (obj, ordering))
            self.ordering[obj] = i
            i += 1

    def __eq__(self, ordering):
        if not isinstance(ordering, ListOrdering):
            return False

        return self.ordering == ordering.ordering

    def in_order(self, x, y):
        return self.cmp(x, y) < 0

    def __contains__(self, var):
        return var in self.ordering

    def cmp(self, x, y):
        return self.ordering[x]-self.ordering[y]

    def get_list(self):
        return sorted(self.ordering.keys(), 
                      key=cmp_to_key((lambda x, y: self.cmp(x, y))))

    def __str__(self):
        return '%s' % (self.get_list())



from ..PL.parser import AST_to_PropositionalLogics
from ..parser import Parser as BaseParser


class AST_to_TemporalLogics(AST_to_PropositionalLogics):
    ''' A class to transform AST into a CTLS temporal formula.

    This class implements a Lark's transformer to turn AST reprensenting
    temporal logics formulas into the corresponding temporal formulas.
    '''

    def __init__(self, lang):
        super(AST_to_TemporalLogics, self).__init__(lang)

    def forall_formula(self, subformulas):
        return self.__lang__.A(*subformulas)

    def exists_formula(self, subformulas):
        return self.__lang__.E(*subformulas)

    def next_formula(self, subformulas):
        return self.__lang__.X(*subformulas)

    def globally_formula(self, subformulas):
        return self.__lang__.G(*subformulas)

    def eventually_formula(self, subformulas):
        return self.__lang__.F(*subformulas)

    def until_formula(self, subformulas):
        return self.__lang__.U(*subformulas)

    def release_formula(self, subformulas):
        return self.__lang__.R(*subformulas)

    def p_formula(self, subformula):
        return subformula[0]


class Parser(BaseParser):
    ''' A class to parse a temporal logics formula.

    Every object of this class instanciates a Lark's parser to parse a string
    reprensenting CTLS formulas into the corresponding pyModelChecking's CTLS
    formulas.
    '''

    grammar = r"""
        s_formula: "{}"     -> true
                 | "{}"    -> false
                 | a_prop
                 | {} u_formula  -> forall_formula
                 | {} u_formula  -> exists_formula
                 | "(" s_formula ")"

        u_formula: {} u_formula  -> next_formula
                 | {} u_formula  -> eventually_formula
                 | {} u_formula  -> globally_formula
                 | {} u_formula  -> not_formula
                 | "(" p_formula ")"
                 | s_formula

        p_formula: u_formula
                 | u_formula ( {} u_formula )+ -> or_formula
                 | u_formula ( {} u_formula )+ -> and_formula
                 | u_formula {} u_formula -> imply_formula
                 | u_formula {} u_formula -> until_formula
                 | u_formula {} u_formula -> release_formula

        a_prop: /[a-zA-Z_][a-zA-Z_0-9]*/ -> string
              | ESCAPED_STRING           -> e_string

        formula: p_formula

        %import common.ESCAPED_STRING
        %import common.WS
        %ignore WS
        """

    def __init__(self, language=None):
        if language is None:
            import pyModelChecking.CTLS as CTLS

            language = CTLS

        super(Parser, self).__init__(grammar=Parser.grammar,
                                     language=language,
                                     AST_Transformer=AST_to_TemporalLogics)


def init_submodule():
    from .language import alphabet

    def format_symbols_for(operator):
        return '({})'.format('|'.join(['\"{}\"'.format(s)
                                       for s in alphabet[operator].symbols]))

    Parser.grammar = Parser.grammar.format(alphabet['Bool'].symbols[True],
                                           alphabet['Bool'].symbols[False],
                                           format_symbols_for('A'),
                                           format_symbols_for('E'),
                                           format_symbols_for('X'),
                                           format_symbols_for('F'),
                                           format_symbols_for('G'),
                                           format_symbols_for('Not'),
                                           format_symbols_for('Or'),
                                           format_symbols_for('And'),
                                           format_symbols_for('Imply'),
                                           format_symbols_for('U'),
                                           format_symbols_for('R'))


init_submodule()



import bisect
import copy
import os
import random
import sys
from collections import deque, namedtuple
import gym
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from rtree import index
import time
import pandas as pd
from trainify.utils.str_list import str_to_list

Transition = namedtuple(
    "Transition", ("state", "action", "reward", "next_state", "done")
)

rindex = '0'
# 获取文件所在的当前路径
script_path = os.path.split(os.path.realpath(__file__))[0]
# 生成需要保存的文件名称以及路径
pt_file = os.path.join(script_path, "dqn" + rindex + ".pt")

class Model(nn.Module):  # 神经网络的类
    def __init__(self, hidden_size, output_size):  # 网络结构及其初始化
        super(Model, self).__init__()
        self.fc1 = nn.Linear(8, hidden_size)
        self.fc1.weight.data.normal_(0, 0.1)
        self.fc1.bias.data.zero_()
        self.fc2 = nn.Linear(hidden_size, hidden_size)
        self.fc2.weight.data.normal_(0, 0.1)
        self.fc2.bias.data.zero_()
        self.fc3 = nn.Linear(hidden_size, hidden_size)
        self.fc3.weight.data.normal_(0, 0.1)
        self.fc3.bias.data.zero_()
        self.fc4 = nn.Linear(hidden_size, output_size)
        self.fc4.weight.data.normal_(0, 0.1)
        self.fc4.bias.data.zero_()

    def forward(self, obs):  # 隐藏层的激活函数
        obs = F.relu(self.fc1(obs))
        obs = F.relu(self.fc2(obs))
        obs = F.relu(self.fc3(obs))
        action_val = self.fc4(obs)
        return action_val
        # 返回的是最终的每个action的数值

class Agent:
    def __init__(self, divide_tool, output_size,
                 hidden_size = 64,
                 memory_capacity = 10000,
                 learning_rate = 1e3,
                 gamma = 0.99,
                 e_greed = 0.1,
                 e_greed_dec = 1e-4,
                 e_greed_min = 0.01,
                 weight_decay = 0,
                 batch_size = 256,
                 warmup_size = 5120,
                 model_sync_count = 8,
                 use_dbqn = False):
        self.model = Model(hidden_size, output_size)  # 当前网络
        self.network = Model(hidden_size, output_size)
        self.target_model = Model(hidden_size, output_size)  # 目标网络
        self.target_model.load_state_dict(self.model.state_dict())  # 当前网络对目标网络赋值
        self.memory = deque(maxlen=memory_capacity)  # 创建双向队列，保存经验回放
        self.optimizer = optim.Adam(
            self.model.parameters(), lr=learning_rate, weight_decay=weight_decay
        )  # 模型的各个参数
        self.loss_func = nn.MSELoss()  # 定义损失函数
        self.e_greed = e_greed
        self.update_count = 0  # 更新次数
        self.noisy = [0, 0, 0, 0]
        self.divide_tool = divide_tool
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.memory_capacity = memory_capacity
        self.learn_rate = learning_rate
        self.weight_decay = weight_decay
        self.e_greed_origin = e_greed
        self.e_greed_dec = e_greed_dec
        self.e_greed_min = e_greed_min
        self.batch_size = batch_size
        self.warmup_size = warmup_size
        self.gamma = gamma
        self.use_dbqn = use_dbqn
        self.model_sync_count = model_sync_count

    def reset(self):
        self.model = Model()  # 当前网络
        self.network = Model()
        self.target_model = Model()  # 目标网络
        self.target_model.load_state_dict(self.model.state_dict())  # 当前网络对目标网络赋值
        self.memory = deque(maxlen=self.memory_capacity)  # 创建双向队列，保存经验回放
        self.optimizer = optim.Adam(
            self.model.parameters(), lr=self.learn_rate, weight_decay=self.weight_decay
        )  # 模型的各个参数
        self.loss_func = nn.MSELoss()  # 定义损失函数
        self.e_greed = self.e_greed_origin
        self.update_count = 0
        self.noisy = [0, 0, 0, 0]

    def update_egreed(self):  # 更新贪心参数e,该函数有问题？？？？
        self.e_greed = max(self.e_greed_min, self.e_greed - self.e_greed_dec)

    def predict(self, obs):  # 选取最大的action
        abs = str_to_list(self.divide_tool.get_abstract_state(obs))
        q_val = self.model(torch.FloatTensor(abs)).detach().numpy()
        q_max = np.max(q_val)
        choice_list = np.where(q_val == q_max)[0]
        return choice_list[0]

    def sample(self, obs):  # 决定是随机选，还是选最大
        if np.random.rand() < self.e_greed:
            return np.random.choice(self.output_size)
        return self.predict(obs)

    def store_transition(self, trans):  # 放入经验回放盒
        self.memory.append(trans)

    def learn(self):  # 该函数实现了：采样，计算损失，反向传播，更新参数
        assert self.warmup_size >= self.batch_size
        if len(self.memory) < self.warmup_size:
            return
        # 从经验回放盒中进行采样
        batch = random.sample(self.memory, self.batch_size)
        batch = Transition(*(zip(*batch)))
        s0 = torch.FloatTensor(batch.state)
        a0 = torch.LongTensor(batch.action).unsqueeze(1)
        r1 = torch.FloatTensor(batch.reward)
        s1 = torch.FloatTensor(batch.next_state)
        d1 = torch.LongTensor(batch.done)

        q_pred = self.model(s0).gather(1, a0).squeeze()
        with torch.no_grad():
            if self.use_dbqn:
                acts = self.model(s1).max(1)[1].unsqueeze(1)
                q_target = self.target_model(s1).gather(1, acts).squeeze(1)
            else:
                q_target = self.target_model(s1).max(1)[0]

            q_target = r1 + self.gamma * (1 - d1) * q_target
        loss = self.loss_func(q_pred, q_target)
        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()
        self.update_count += 1
        if self.update_count % self.model_sync_count == 0:
            self.target_model.load_state_dict(self.model.state_dict())

    def save(self, sava_path = pt_file):  # 保存网络的参数数据
        torch.save(self.model.state_dict(), sava_path)
        # print(pt_file + " saved.")

    def load(self, load_path = pt_file):  # 加载网络的参数数据
        self.model.load_state_dict(torch.load(load_path))
        self.network.load_state_dict(torch.load(load_path))
        self.target_model.load_state_dict(self.model.state_dict())
        print(load_path + " loaded.")

    def add_noisy(self, obs):

        obs[0] += random.gauss(0, self.noisy[0])
        obs[1] += random.gauss(0, self.noisy[1])
        obs[2] += random.gauss(0, self.noisy[2])
        obs[3] += random.gauss(0, self.noisy[3])
        return obs

    def update_noisy(self):
        # noisy = [0.005, 0.01, 0.0005, 0.01]
        step_length = [0.005, 0.005, 0.005, 0.005]
        for i in range(len(step_length)):
            self.noisy[i] += step_length[i]




def clip(state):
    # low = state_space[0][0]
    # high = state_space[1][0]
    state_space1 = [[-4.79, -9.99, -0.419, -9.99], [4.79, 9.99, 0.419, 9.99]]
    state[0] = np.clip(state[0], state_space1[0][0], state_space1[1][0])
    state[1] = np.clip(state[1], state_space1[0][1], state_space1[1][1])
    state[2] = np.clip(state[2], state_space1[0][2], state_space1[1][2])
    state[3] = np.clip(state[3], state_space1[0][3], state_space1[1][3])
    return state




import requests
import json

_APPKEY = "BC-c745270abeb04089a10632bbb6bec81b"
_GOEASY_URL = "http://rest-hangzhou.goeasy.io/publish"


def send_to_backend(channel, content):
    data = {"appkey": _APPKEY, "channel": channel, "content": json.dumps(content)}
    # print(data)
    return requests.post(_GOEASY_URL, data=data)



def str_to_list(state_str):
    return list(map(float, state_str.split(',')))


def list_to_str(state_list):
    res_list = []
    for tmp in state_list:
        obj_str = ','.join([str(_) for _ in tmp])
        res_list.append(obj_str)
    return res_list



"""
.. module:: CTLS.model_checking
   :synopsis: Provides model checking methods for the CTL* language.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from trainify.validator.pyModelChecking.kripke import Kripke

from .parser import Parser

import sys

import trainify.validator.pyModelChecking.CTL
import trainify.validator.pyModelChecking.LTL

CTL = sys.modules['trainify.validator.pyModelChecking.CTL']
LTL = sys.modules['trainify.validator.pyModelChecking.LTL']


def _get_a_new_atomic_proposition_for(kripke, formula):
    f_str = '[{}]'.format(formula)
    f_atom = f_str
    atoms = kripke.labels()

    i = 0
    while f_atom in atoms:
        f_atom = '[{}({})]'.format(f_str, i)
        i += 1

    return f_atom


def _remove_state_subformulas(kripke, formula, fair_label=None):
    Lang = sys.modules[formula.__module__]

    if isinstance(formula, AtomicProposition):
        return formula

    if isinstance(formula, PathQuantifier):
        f_atom = _get_a_new_atomic_proposition_for(kripke, formula)
        for s in _checkQuantifiedFormula(kripke, formula, fair_label):
            kripke.labels(s).add(f_atom)

        return Lang.AtomicProposition(f_atom)

    if isinstance(formula, Formula):
        sfs = []
        for sf in formula.subformulas():
            sfs.append(_remove_state_subformulas(kripke, sf, fair_label))

        return formula.__class__(*sfs)

    raise TypeError('expected a CTL* state formula, got {}' % (formula))


def _checkQuantifiedFormula(kripke, formula, fair_label=None):
    subformula = _remove_state_subformulas(kripke, formula.subformula(0),
                                           fair_label)

    formula = formula.__class__(subformula)
    try:
        if fair_label is not None:
            formula = formula.cast_to(CTL)
            formula = (formula).get_equivalent_non_fair_formula(fair_label)

        return CTL.modelcheck(kripke, formula)
    except TypeError:
        if fair_label is not None:
            formula = formula.get_equivalent_non_fair_formula(fair_label)

        if (not isinstance(formula, A)):
            if (isinstance(formula, E)):
                formula = LNot(A(LNot(formula.subformula(0))))

            formula = _remove_state_subformulas(kripke, formula)

            return CTL.modelcheck(kripke, formula)

        return LTL.modelcheck(kripke, formula)


def modelcheck(kripke, formula, parser=None, F=None):
    ''' Model checks any CTL* formula on a Kripke structure.

    This method performs CTL* model checking of a formula on a given
    Kripke structure.

    :param kripke: a Kripke structure.
    :type kripke: Kripke
    :param formula: the formula to model check.
    :type formula: a type castable in a CTLS.Formula or a string representing
                   a CTLS formula
    :param parser: a parser to parse a string into a CTLS.Formula.
    :type parser: CTLS.Parser
    :param F: a list of fair states
    :type F: Container
    :returns: a list of the Kripke structure states that satisfy the formula.
    '''

    if isinstance(formula, str):
        if parser is None:
            parser = Parser()
        formula = parser(formula)

    if not isinstance(kripke, Kripke):
        raise TypeError('expected a Kripke structure, got {}'.format(kripke))

    try:
        kripkeC = kripke.clone()

        if F is not None:
            fair_label = kripkeC.label_fair_states(F)

            CTL_frml = _remove_state_subformulas(kripkeC, formula,
                                                 fair_label)

            CTL_frml = CTL_frml.get_equivalent_non_fair_formula(fair_label)
        else:
            CTL_frml = _remove_state_subformulas(kripkeC, formula)

        return CTL.modelcheck(kripkeC, CTL_frml)

    except TypeError as e:
        print(e)

        raise TypeError('expected a CTL* state formula, ' +
                        'got {}'.format(formula))



from ..parser import Parser as BaseParser
from ..CTLS.parser import AST_to_TemporalLogics


class Parser(BaseParser):
    grammar = r"""
        s_formula: "{}"     -> true
                 | "{}"    -> false
                 | a_prop
                 | {} p_formula       -> forall_formula
                 | {} p_formula       -> exists_formula
                 | {} s_formula       -> not_formula
                 | "(" u_formula ")"

        u_formula: s_formula
                  | s_formula ( {} s_formula )+      -> or_formula
                  | s_formula ( {} s_formula )+      -> and_formula
                  | s_formula {} s_formula           -> imply_formula

        p_formula: {} s_formula  -> next_formula
                 | {} s_formula  -> eventually_formula
                 | {} s_formula  -> globally_formula
                 | s_formula {} s_formula -> until_formula
                 | s_formula {} s_formula -> release_formula
                 | "(" p_formula ")"

        a_prop: /[a-zA-Z_][a-zA-Z_0-9]*/ -> string
              | ESCAPED_STRING           -> e_string

        formula: p_formula | u_formula

        %import common.ESCAPED_STRING
        %import common.WS
        %ignore WS
        """

    def __init__(self, language=None):
        if language is None:
            import trainify.validator.pyModelChecking.CTL as CTL

            language = CTL

        super(Parser, self).__init__(grammar=Parser.grammar,
                                     language=language,
                                     AST_Transformer=AST_to_TemporalLogics)


def init_submodule():
    from .language import alphabet

    def format_symbols_for(operator):
        return '({})'.format('|'.join(['\"{}\"'.format(s)
                                       for s in alphabet[operator].symbols]))

    Parser.grammar = Parser.grammar.format(alphabet['Bool'].symbols[True],
                                           alphabet['Bool'].symbols[False],
                                           format_symbols_for('A'),
                                           format_symbols_for('E'),
                                           format_symbols_for('Not'),
                                           format_symbols_for('Or'),
                                           format_symbols_for('And'),
                                           format_symbols_for('Imply'),
                                           format_symbols_for('X'),
                                           format_symbols_for('F'),
                                           format_symbols_for('G'),
                                           format_symbols_for('U'),
                                           format_symbols_for('R'))


init_submodule()



import gym
from gym import spaces
from gym.utils import seeding
import numpy as np


class B4Env(gym.Env):
    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 30
    }

    def __init__(self):
        self.th = 1.0
        # self.viewer = None

        high = np.array([2, 2, 2], dtype=np.float32)
        self.action_space = spaces.Box(
            low=-self.th,
            high=self.th,
            shape=(1,),
            dtype=np.float32
        )
        self.observation_space = spaces.Box(
            low=-high,
            high=high,
            dtype=np.float32
        )

        self.seed()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, u):
        x1, x2, x3 = self.state
        done = False
        # u = np.clip(u, -self.th, self.th)[0]
        # u = 10 * u

        offset = 0
        scala = 1

        # offset = 0.5
        # scala = 2

        u = u[0] - offset
        u = u * scala
        # t = 0.02
        t = 0.05
        x1_new = x1 + (-x1 + x2 - x3) * t
        x2_new = x2 + (-x1 * (x3 + 1) - x2) * t
        x3_new = x3 + (-x1 + u) * t

        self.state = np.array([x1_new, x2_new, x3_new], dtype=np.float32)

        reward = (-abs(x1_new) - abs(x2_new + 0.03)) * 2

        # if 0.05 >= x1_new >= -0.05 and 0 >= x2_new >= -0.05:
        #     reward = 500
        #     done = True
        if 0.04 >= x1_new >= -0.04 and -0.014 >= x2_new >= -0.045:
            reward = 500
            done = True

        # done = bool(
        #     abs(x1_new) > 1.5 or
        #     abs(x2_new) > 1.5
        # ) or done
        #
        # if bool(
        #         abs(x1_new) > 1.5 or
        #         abs(x2_new) > 1.5
        # ):
        #     reward = -600

        return self._get_obs(), reward, done, {}

    def reset(self):
        high = np.array([0.27, 0.1, 0.27])
        low = np.array([0.25, 0.08, 0.25])
        self.state = self.np_random.uniform(low=low, high=high)

        return self._get_obs()

    def _get_obs(self):
        self.state[0] = np.clip(self.state[0], -2, 2)
        self.state[1] = np.clip(self.state[1], -2, 2)
        self.state[2] = np.clip(self.state[2], -2, 2)
        return self.state
        # return np.array(transition([np.cos(theta), np.sin(theta), thetadot]))

    def step_size(self, u, step_size=0.005):

        done = False

        offset = 0
        scala = 1
        u = u[0] - offset
        u = u * scala
        t = 0.1
        time = 0
        state_list = []
        while time <= t:
            x1, x2, x3 = self.state
            x1_new = x1 + (-x1 + x2 - x3) * step_size
            x2_new = x2 + (-x1 * (x3 + 1) - x2) * step_size
            x3_new = x3 + (-x1 + u) * step_size
            state_list.append([x1_new, x2_new])

            self.state = np.array([x1_new, x2_new, x3_new], dtype=np.float32)

            if 0.05 >= x1_new >= -0.05 and 0 >= x2_new >= -0.05:
                done = True
            time = round(time + step_size, 10)
        return self.state, state_list, done



import time
import logging
import sys
import colorlog
import os
from trainify.data.backend_logger_handler import BackendHandler

# 是否开启log日志
LOG_ENABLED = True
# 是否输出到控制台
LOG_TO_CONSOLE = True
# 是否输出到文件
LOG_TO_FILE = True
# 日志等级
LOG_LEVEL = logging.DEBUG
# 每条日志输出格式
LOG_FILE_FORMAT = '[%(asctime)s.%(msecs)03d] %(filename)s -> %(funcName)s line:%(lineno)d [%(levelname)s] : %(message)s'
LOG_CONSOLE_FORMAT = '%(log_color)s[%(asctime)s.%(msecs)03d] %(filename)s -> %(funcName)s line:%(lineno)d [%(' \
                     'levelname)s] : %(message)s '
LOG_CONSOLE_COLOR = {
    'DEBUG': 'white',  # cyan white
    'INFO': 'green',
    'WARNING': 'yellow',
    'ERROR': 'red',
    'CRITICAL': 'bold_red',
}


class Logger(object):
    def __init__(self, log_path, channel):
        self.loggers = {}
        self.enabled = LOG_ENABLED  # 是否开启日志
        self.is_console = LOG_TO_CONSOLE  # 是否输出到控制台
        self.is_file = LOG_TO_FILE  # 是否输出到文件
        self.path = log_path  # 日志文件路径
        self.level = LOG_LEVEL  # 日志级别
        # self.format = LOG_FORMAT  # 每条日志输出格式
        self.backup_count = 1000000  # 最多存放日志的数量
        self.channel = channel

    def create_logger(self, logger_name=None):
        if logger_name is None:
            logger_name = 'log_' + time.strftime("_%Y%m%d_%H%M%S", time.localtime())

        if not os.path.exists(self.path):
            os.makedirs(self.path)
        print(logger_name)
        logger = logging.getLogger(logger_name)
        logger.setLevel(self.level)
        if self.enabled and self.is_console:
            console_handler = logging.StreamHandler(sys.stdout)
            # console_handler.setLevel(self.level)
            console_formatter = colorlog.ColoredFormatter(
                fmt=LOG_CONSOLE_FORMAT,
                datefmt='%Y-%m-%d  %H:%M:%S',
                log_colors=LOG_CONSOLE_COLOR
            )
            console_handler.setFormatter(console_formatter)
            logger.addHandler(console_handler)

        if self.enabled and self.is_file:
            file_handler = logging.FileHandler(self.path + '/' + logger_name + '.log')
            # file_handler.setLevel(self.level)
            file_handler.setFormatter(logging.Formatter(LOG_FILE_FORMAT))
            logger.addHandler(file_handler)

        if self.enabled and self.channel:
            back_handler = BackendHandler(self.channel)
            back_handler.setFormatter(logging.Formatter(LOG_FILE_FORMAT))
            logger.addHandler(back_handler)
        return logger



from lark import Lark, Transformer, exceptions


class AST_to_Logics(Transformer, object):
    ''' A class to transform AST into a propositional formula.

    This class implements a Lark's transformer to turn AST reprensenting
    propositional logics formulas into the corresponding propositional formula.
    '''

    def __init__(self, lang):
        super(AST_to_Logics, self).__init__()

        self.__lang__ = lang

    def or_formula(self, subformulas):
        return self.__lang__.Or(*subformulas)

    def and_formula(self, subformulas):
        return self.__lang__.And(*subformulas)

    def imply_formula(self, subformulas):
        return self.__lang__.Imply(*subformulas)

    def not_formula(self, subformulas):
        return self.__lang__.Not(*subformulas)

    def true(self, subformula):
        return self.__lang__.Bool(True)

    def false(self, subformula):
        return self.__lang__.Bool(False)


class ParserError(Exception):
    ''' A class to represent parser errors. '''

    def __init__(self, string, pos):
        self.string = string
        self.pos = pos

    def __str__(self):
        return 'Parser Error: \n\n{}\n{}^\n'.format(self.string, ' '*self.pos)


class UnexpectedToken(ParserError):
    def __init__(self, string, pos):
        super(UnexpectedToken, self).__init__(string, pos)


class UnexpectedCharacters(ParserError):
    def __init__(self, string, pos):
        super(UnexpectedCharacters, self).__init__(string, pos)


class Parser(object):
    ''' A class to parse a propositional formula.

    Every object of this class instanciates a Lark's parser to parse a
    string reprensenting propositional formulas into the corresponding
    pyModelChecking's formulas.
    '''

    def __init__(self, grammar, language, AST_Transformer):
        self._parser = Lark(grammar, start='formula', parser='lalr',
                            transformer=AST_Transformer(language))

    def __call__(self, string):
        ''' Parses a string and returns the equivalent formula

        This method parses a string and returns the equivalent formula.

        :param name: a string representing a formula.
        :type name: str
        :returns: a formula.
        :rtype: a temporal formula type
        :raises: objects of either :py:exception:UnexpectedToken or
                 :py:exception:UnexpectedCharacters
        '''
        try:
            return self._parser.parse(string)
        except exceptions.UnexpectedToken as e:
            ex_class = UnexpectedToken
            pos = int(e.pos_in_stream)
        except exceptions.UnexpectedCharacters as e:
            ex_class = UnexpectedCharacters
            pos = int(e.pos_in_stream)

        raise ex_class(string, pos)



"""
.. module:: CTL
   :synopsis: Represents Binary Decision Diagrams.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .BDD import BDDNode
from .ordering import Ordering
from .OBDD import OBDD



import gym
from gym import spaces
from gym.utils import seeding
import numpy as np


class B1Env(gym.Env):
    metadata = {
        'render.modes': ['human', 'rgb_array'],
        'video.frames_per_second': 30
    }

    def __init__(self):
        self.th = 1.0
        # self.viewer = None

        high = np.array([5.0, 5.0], dtype=np.float32)
        self.action_space = spaces.Box(
            low=-self.th,
            high=self.th,
            shape=(1,),
            dtype=np.float32
        )
        self.observation_space = spaces.Box(
            low=-high,
            high=high,
            dtype=np.float32
        )

        self.seed()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def step(self, u):
        p, v = self.state
        done = False
        # u = np.clip(u, -self.th, self.th)[0]
        # offset = 0.1
        offset = 0
        # scala = 4
        scala = 1
        u = u[0] - offset
        u = scala * u
        t = 0.1
        p_new = p + v * t
        v_new = v + (u * v * v - p) * t

        self.state = np.array([p_new, v_new], dtype=np.float32)

        reward = -2

        # if 0.2 >= p_new >= 0 and 0.3 >= v_new >= 0.05:
        #     # reward += 100
        #     done = True
        if 0.15 >= p_new >= 0.02 and 0.28 >= v_new >= 0.07:
            # reward += 100
            done = True
        # if 0.15 >= p_new >= 0.05 and 0.25 >= v_new >= 0.1:
        #     # reward += 100
        #     done = True

        done = bool(
            abs(p_new) > 1.5 or
            abs(v_new) > 1.5
        ) or done

        if bool(
                abs(p_new) > 1.5 or
                abs(v_new) > 1.5
        ):
            reward = -600

        return self._get_obs(), reward, done, {}

    def reset(self):
        high = np.array([0.9, 0.6])
        low = np.array([0.8, 0.5])
        self.state = self.np_random.uniform(low=low, high=high)

        return self._get_obs()

    def _get_obs(self):
        self.state[0] = np.clip(self.state[0], -2, 2)
        self.state[1] = np.clip(self.state[1], -2, 2)
        return self.state
        # return np.array(transition([np.cos(theta), np.sin(theta), thetadot]))

    def step_size(self, u, step_size=0.001):
        t = 0.1
        time = 0
        state_list = []
        done = False
        offset = 0
        scala = 4
        u = u[0] - offset
        u = scala * u
        while time <= t:
            p, v = self.state
            # u = np.clip(u, -self.th, self.th)[0]
            # offset = 0.1
            p_new = p + v * step_size
            v_new = v + (u * v * v - p) * step_size
            self.state = np.array([p_new, v_new], dtype=np.float32)
            state_list.append([p_new, v_new])
            if 0.15 >= p_new >= 0.02 and 0.28 >= v_new >= 0.07:
                # reward += 100
                done = True
            time = round(time + step_size, 10)
        return self.state, state_list, done

# def angle_normalize(self, x):
#     return (( (x + np.pi) % (2 * np.pi) ) - np.pi)



from pyModelChecking.kripke import *
import unittest


class TestKripke(unittest.TestCase):

    def setUp(self):
        self.S = set([0, 1, 3])
        self.S0 = set([0, 1, 6])
        self.R = set([(0, 2), (2, 2), (0, 1), (1, 0), (3, 2)])

        self.L = {1: set(['p', 'q']), 2: set(['p', 'q']), 3: set(['q'])}

        self.K = Kripke(self.S, self.S0, self.R, self.L)

    def test_init(self):
        S = set(self.S)
        for (s, d) in self.R:
            S.add(s)
            S.add(d)

        self.assertEqual(set(self.K.states()), S)

        with self.assertRaises(RuntimeError):
            Kripke(self.S, self.S0, self.R | set([(2, 4)]), self.L)

    def test_nodes(self):
        S = self.S | set([2])

        self.assertEqual(set(self.K.states()), S)

    def test_edges(self):
        self.assertEqual(set(self.K.transitions()), self.R)

    def test_labels(self):
        AP = set()
        for ap in self.L.values():
            AP.update(ap)

        self.assertEqual(self.K.labels(), AP)

        for v in set(self.K.states())-set(self.L.keys()):
            self.assertEqual(self.K.labels(v), set())

        for v in set(self.L.keys()):
            self.assertEqual(self.K.labels(v), self.L[v])

        with self.assertRaises(RuntimeError):
            self.K.labels('a')

    def test_next(self):

        for s in self.K.nodes():
            N = set()
            for (p, q) in self.K.edges_iter():
                if s == p:
                    N.add(q)

            self.assertEqual(self.K.next(s), N)

        with self.assertRaises(RuntimeError):
            self.K.next('a')


if __name__ == '__main__':
    unittest.main()



"""
.. module:: CTLS
   :synopsis: Represents the CTL* language and provides model
              checking methods for it.

.. moduleauthor:: Alberto Casagrande <acasagrande@units.it>
"""

from .language import *
from .model_checking import modelcheck

from ..language import LNot

from .parser import Parser



from pyModelChecking.LTL import *
from pyModelChecking.parser import UnexpectedToken, UnexpectedCharacters

import unittest
import lark


class TestLTLParser(unittest.TestCase):
    def setUp(self):
        self.same_formulas = [(Bool(True), "true"),
                              (Bool(False), "false"),
                              (Not(True), "not true"),
                              (A(F(G('q'))), "A F G q"),
                              (A(Imply(F(G('q')), X('p'))),
                               "A (F G q --> X p)"),
                              (A(G(Imply(And(Not('Close'), 'Start'),
                                         Or(G(Not('Heat')),
                                            F(Not('Error')))))),
                               "A G ((not Close and Start) --> " +
                               "((G not Heat) or (F not Error)))")]

        self.wrong_formulas = [("p and", UnexpectedToken),
                               ("A U and G", UnexpectedToken),
                               ("A a --> +", UnexpectedToken),
                               ("true\nfalse", UnexpectedToken),
                               ("A (F G q --> E G p)", UnexpectedToken),
                               ("A G ((not Close and Start) --> " +
                                "A ((G not Heat) or (F not Error)))",
                                UnexpectedToken)]

    def test_parser(self):
        parser = Parser()
        for psi, phi_str in self.same_formulas:
            phi = parser(phi_str)

            self.assertEqual(phi, psi)

        for phi_str, ErrorType in self.wrong_formulas:
            with self.assertRaises(ErrorType):
                phi = parser(phi_str)


if __name__ == '__main__':
    unittest.main()



import time

import numpy as np
from rtree import index

from trainify.validator import FMValidator

record_num = 1


def cegar(rtree_name, agent, divide_tool, train_func, verify_env, config, recorder):
    max_iteration = config['max_iteration']

    recorder.logger.info('Verify cegar验证 开始时间 ' + time.strftime("%Y-%m-%d-%H_%M_%S", time.localtime()))
    t0 = time.time()
    # 此时agent应该已经训练好了
    # train_model(agent)
    # agent.load()
    # evaluate(agent)
    tr = time.time()
    v = FMValidator(verify_env, agent.network, recorder.logger)
    k = v.create_kripke_ctl()
    t1 = time.time()
    if k is None:
        violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
        res2 = True
    else:
        # v.formula = 'not(A(G(safe)))'
        res2, violated_states = v.ctl_model_check(k)
    recorder.logger.info('Verify cegar number of counterexamples：' + str(len(violated_states)))
    t2 = time.time()
    recorder.logger.info('Verify cegar train time: ' + str(tr - t0) + ' construct kripke structure: ' + str(
        t1 - tr) + ' model checking: ' + str(t2 - t1))
    recorder.logger.info(time.strftime("%Y-%m-%d-%H_%M_%S " + str(time.localtime())))
    start_id = divide_tool.rtree.get_size()
    p = index.Property()
    p.dimension = len(divide_tool.key_dim)
    iteration_time = 1
    while res2 and iteration_time <= max_iteration:
        recorder.logger.info(time.strftime("%Y-%m-%d-%H_%M_%S " + str(time.localtime())))
        t0 = time.time()
        divide_tool.rtree = index.Index(rtree_name, divide_tool.rtree_refinement(violated_states, start_id, recorder),
                                        properties=p)
        recorder.logger.info('Verify cegar number of states after refinement: ' + str(divide_tool.rtree.get_size()))

        recorder.add_chart('精化后的抽象状态数量', {
            "desc": "精化后的抽象状态数量",
            'y': len(divide_tool.rtree),
        })
        trr = time.time()
        start_id += divide_tool.rtree.get_size()
        agent.divide_tool = divide_tool
        # agent.load()
        agent.reset()
        train_func()

        agent.load_model()
        # evaluate(agent)
        tr = time.time()
        verify_env.divide_tool = divide_tool
        verify_env.network = agent.network
        # pd = PendulumEnv(divide_tool, agent.actor)
        v = FMValidator(verify_env, agent.network, recorder.logger)
        k = v.create_kripke_ctl()
        t1 = time.time()
        if k is None:
            violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
            res2 = True
        else:
            # v.formula = 'not(A(G(safe)))'
            res2, violated_states = v.ctl_model_check(k)
        recorder.logger.info('Verify cegar number of counterexamples：' + str(len(violated_states)))
        recorder.add_chart('反例数量', {
            "desc": "验证过程中的反例数量",
            'y': len(violated_states),
        })
        recorder.logger.info('Verify cegar iteration :' + str(iteration_time) + str(res2) + str(len(violated_states)))
        t2 = time.time()
        recorder.logger.info('Verify cegar refine: ' + str(trr - t0) + str(' train:') + str(
            tr - trr) + 'construct kripke structure: ' + str(t1 - tr) + ' model checking: ' + str(t2 - t1))
        recorder.logger.info(time.strftime("%Y-%m-%d-%H_%M_%S" + str(time.localtime())))
        iteration_time += 1
    while res2:
        t0 = time.time()
        # agent.load()
        agent.reset()
        train_func()

        agent.load_model()
        tr = time.time()
        # verify_env.divide_tool = divide_tool
        verify_env.network = agent.network
        v = FMValidator(verify_env, agent.network, recorder.logger)
        k = v.create_kripke_ctl()
        t1 = time.time()
        if k is None:
            violated_states = list(divide_tool.rtree.intersection(divide_tool.rtree.bounds, objects='raw'))
            res2 = True
            recorder.logger.info('Verify cegar number of counterexamples：' + str(len(violated_states)))
        else:
            # v.formula = 'not(A(G(safe)))'
            res2, violated_states = v.ctl_model_check(k)
            recorder.logger.info('Verify cegar number of counterexamples：' + str(len(violated_states)))
        t2 = time.time()
        recorder.logger.info('Verify cegar train: ' + str(tr - t0) +
                             'construct kripke structure: ' + str(t1 - tr) + str(' model checking: ') + str(t2 - t1))
        recorder.logger.info(time.strftime("%Y-%m-%d-%H_%M_%S" + str(time.localtime())))

    recorder.logger.info(
        'Verify 验证结束 ' + ' 构建kripke所花时间: ' + str(t1 - tr) + str(' 模型检查时间: ') + str(t2 - t1))



from ..parser import Parser as BaseParser
from ..CTLS.parser import AST_to_TemporalLogics


class Parser(BaseParser):
    grammar = r"""
        formula: s_formula | p_formula

        s_formula: {} u_formula      -> forall_formula

        p_formula: u_formula ( {} u_formula )+   -> or_formula
                 | u_formula ( {} u_formula )+   -> and_formula
                 | u_formula {} u_formula        -> imply_formula
                 | u_formula {} u_formula        -> until_formula
                 | u_formula {} u_formula        -> release_formula
                 | u_formula

        u_formula: "{}"                -> true
                 | "{}"                -> false
                 | a_prop
                 | "(" p_formula ")"
                 | {} u_formula      -> not_formula
                 | {} u_formula      -> next_formula
                 | {} u_formula      -> eventually_formula
                 | {} u_formula      -> globally_formula

        a_prop: /[a-zA-Z_][a-zA-Z_0-9]*/ -> string
              | ESCAPED_STRING           -> e_string

        %import common.ESCAPED_STRING
        %import common.WS
        %ignore WS
        """

    def __init__(self, language=None):
        if language is None:
            import pyModelChecking.LTL as LTL

            language = LTL

        super(Parser, self).__init__(grammar=Parser.grammar,
                                     language=language,
                                     AST_Transformer=AST_to_TemporalLogics)


def init_submodule():
    from .language import alphabet

    def format_symbols_for(operator):
        return '({})'.format('|'.join(['\"{}\"'.format(s)
                                       for s in alphabet[operator].symbols]))

    Parser.grammar = Parser.grammar.format(format_symbols_for('A'),
                                           format_symbols_for('Or'),
                                           format_symbols_for('And'),
                                           format_symbols_for('Imply'),
                                           format_symbols_for('U'),
                                           format_symbols_for('R'),
                                           alphabet['Bool'].symbols[True],
                                           alphabet['Bool'].symbols[False],
                                           format_symbols_for('Not'),
                                           format_symbols_for('X'),
                                           format_symbols_for('F'),
                                           format_symbols_for('G'))


init_submodule()
